/* hp2100_stddev.c: HP2100 standard devices simulator

*****
***** Modified 7/12/08 by WTN...
***** HP2615/21/45 terminal emulation (sort of) by converting
***** common sequences into standard VT-type escape sequences.
***** This file is for v3.8-0, if using another version merge
***** the changes into the existing source.
***** SET TTY1 TRANSLATE to enable
***** SET TTY1 NOTRANSLATE to disable (default)
***** (can be set for TTY0/2 but no effect)
*****

   Copyright (c) 1993-2008, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not be
   used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   ptr          12597A-002 paper tape reader interface
   ptp          12597A-005 paper tape punch interface
   tty          12531C buffered teleprinter interface
   clk          12539C time base generator

   25-Apr-08    JDB     Changed TTY output wait from 100 to 200 for MSU BASIC
   18-Apr-08    JDB     Removed redundant control char handling definitions
   14-Apr-08    JDB     Changed TTY console poll to 10 msec. real time
                        Synchronized CLK with TTY if set for 10 msec.
                        Added UNIT_IDLE to TTY and CLK
   09-Jan-08    JDB     Fixed PTR trailing null counter for tape re-read
   31-Dec-07    JDB     Added IPTICK register to CLK to display CPU instr/tick
                        Corrected and verified ioCRS actions
   28-Dec-06    JDB     Added ioCRS state to I/O decoders
   22-Nov-05    RMS     Revised for new terminal processing routines
   13-Sep-04    JDB     Added paper tape loop mode, DIAG/READER modifiers to PTR
                        Added PV_LEFT to PTR TRLLIM register
                        Modified CLK to permit disable
   15-Aug-04    RMS     Added tab to control char set (from Dave Bryan)
   14-Jul-04    RMS     Generalized handling of control char echoing
                        (from Dave Bryan)
   26-Apr-04    RMS     Fixed SFS x,C and SFC x,C
                        Fixed SR setting in IBL
                        Fixed input behavior during typeout for RTE-IV
                        Suppressed nulls on TTY output for RTE-IV
                        Implemented DMA SRQ (follows FLG)
   29-Mar-03    RMS     Added support for console backpressure
   25-Apr-03    RMS     Added extended file support
   22-Dec-02    RMS     Added break support
   01-Nov-02    RMS     Revised BOOT command for IBL ROMs
                        Fixed bug in TTY reset, TTY starts in input mode
                        Fixed bug in TTY mode OTA, stores data as well
                        Fixed clock to add calibration, proper start/stop
                        Added UC option to TTY output
   30-May-02    RMS     Widened POS to 32b
   22-Mar-02    RMS     Revised for dynamically allocated memory
   03-Nov-01    RMS     Changed DEVNO to use extended SET/SHOW
   29-Nov-01    RMS     Added read only unit support
   24-Nov-01    RMS     Changed TIME to an array
   07-Sep-01    RMS     Moved function prototypes
   21-Nov-00    RMS     Fixed flag, buffer power up state
                        Added status input for ptp, tty
   15-Oct-00    RMS     Added dynamic device number support

   The reader and punch, like most HP devices, have a command flop.  The
   teleprinter and clock do not.

   Reader diagnostic mode simulates a tape loop by rewinding the tape image file
   upon EOF.  Normal mode EOF action is to supply TRLLIM nulls and then either
   return SCPE_IOERR or SCPE_OK without setting the device flag.

   To support CPU idling, the teleprinter interface (which doubles as the
   simulator console) polls for input using a calibrated timer with a ten
   millisecond period.  Other polled-keyboard input devices (multiplexers and
   the BACI card) synchronize with the console poll to ensure maximum available
   idle time.  The console poll is guaranteed to run, as the TTY device cannot
   be disabled.

   The clock (time base generator) autocalibrates.  If the CLK is set to a ten
   millisecond period (e.g., as under RTE), it is synchronized to the console
   poll.  Otherwise (e.g., as under DOS or TSB, which use 100 millisecond
   periods), it runs asynchronously.  If the specified clock frequency is below
   10Hz, the clock service routine runs at 10Hz and counts down a repeat counter
   before generating an interrupt.  Autocalibration will not work if the clock
   is running at 1Hz or less.

   Clock diagnostic mode corresponds to inserting jumper W2 on the 12539C.
   This turns off autocalibration and divides the longest time intervals down
   by 10**3.  The clk_time values were chosen to allow the diagnostic to
   pass its clock calibration test.

   References:
   - 2748B Tape Reader Operating and Service Manual (02748-90041, Oct-1977)
   - 12597A 8-Bit Duplex Register Interface Kit Operating and Service Manual
            (12597-9002, Sep-1974)
   - 12531C Buffered Teleprinter Interface Kit Operating and Service Manual
            (12531-90033, Nov-1972)
   - 12539C Time Base Generator Interface Kit Operating and Service Manual
            (12539-90008, Jan-1975)
*/

#include "hp2100_defs.h"

#define TTY_OUT_WAIT    200                             /* TTY output wait */

#define UNIT_V_DIAG     (TTUF_V_UF + 0)                 /* diag mode */
#define UNIT_V_AUTOLF   (TTUF_V_UF + 1)                 /* auto linefeed */
#define UNIT_DIAG       (1 << UNIT_V_DIAG)
#define UNIT_AUTOLF     (1 << UNIT_V_AUTOLF)

#define PTP_LOW         0000040                         /* low tape */
#define TM_MODE         0100000                         /* mode change */
#define TM_KBD          0040000                         /* enable keyboard */
#define TM_PRI          0020000                         /* enable printer */
#define TM_PUN          0010000                         /* enable punch */
#define TP_BUSY         0100000                         /* busy */

#define CLK_V_ERROR     4                               /* clock overrun */
#define CLK_ERROR       (1 << CLK_V_ERROR)

/* added by WTN */
#define UNIT_V_TRANSLATE (TTUF_V_UF + 2)
#define UNIT_TRANSLATE   (1 << UNIT_V_TRANSLATE)
int32 escflag = 0;  /* flag for detecting escapes         */
int32 escstate = 0; /* var for sequencing next action     */
int32 escchar1 = 0; /* 1st char of numerical control      */
int32 escchar2 = 0; /* 2nd char of numerical control      */ 
int32 escchar3 = 0; /* 3rd char of numerical control      */ 
int32 pnormal = 1;  /* 0=don't print, 1=print normally    */
int32 cancelesc = '~'; /* for cancelling escape sequences */
int32 doingattrib = 0; /* for cancelling attribs on crlf  */
/* end WTN vars */

extern uint32 PC, SR;
extern uint32 dev_cmd[2], dev_ctl[2], dev_flg[2], dev_fbf[2], dev_srq[2];

int32 ptr_stopioe = 0;                                  /* stop on error */
int32 ptr_trlcnt = 0;                                   /* trailer counter */
int32 ptr_trllim = 40;                                  /* trailer to add */
int32 ptp_stopioe = 0;
int32 ttp_stopioe = 0;
int32 tty_buf = 0;                                      /* tty buffer */
int32 tty_mode = 0;                                     /* tty mode */
int32 tty_shin = 0377;                                  /* tty shift in */
int32 tty_lf = 0;                                       /* lf flag */
int32 clk_select = 0;                                   /* clock time select */
int32 clk_error = 0;                                    /* clock error */
int32 clk_ctr = 0;                                      /* clock counter */
int32 clk_time[8] = {                                   /* clock intervals */
    155, 1550, 15500, 155000, 155000, 155000, 155000, 155000
    };
int32 clk_tps[8] = {                                    /* clock tps */
    10000, 1000, 100, 10, 10, 10, 10, 10
    };
int32 clk_rpt[8] = {                                    /* number of repeats */
    1, 1, 1, 1, 10, 100, 1000, 10000
    };
uint32 clk_tick = 0;                                    /* instructions per tick */

DEVICE ptr_dev, ptp_dev, tty_dev, clk_dev;
int32 ptrio (int32 inst, int32 IR, int32 dat);
t_stat ptr_svc (UNIT *uptr);
t_stat ptr_attach (UNIT *uptr, char *cptr);
t_stat ptr_reset (DEVICE *dptr);
t_stat ptr_boot (int32 unitno, DEVICE *dptr);
int32 ptpio (int32 inst, int32 IR, int32 dat);
t_stat ptp_svc (UNIT *uptr);
t_stat ptp_reset (DEVICE *dptr);
int32 ttyio (int32 inst, int32 IR, int32 dat);
t_stat tti_svc (UNIT *uptr);
t_stat tto_svc (UNIT *uptr);
t_stat tty_reset (DEVICE *dptr);
t_stat tty_set_opt (UNIT *uptr, int32 val, char *cptr, void *desc);
t_stat tty_set_alf (UNIT *uptr, int32 val, char *cptr, void *desc);
int32 clkio (int32 inst, int32 IR, int32 dat);
t_stat clk_svc (UNIT *uptr);
t_stat clk_reset (DEVICE *dptr);
int32 clk_delay (int32 flg);
t_stat tto_out (int32 c);
t_stat ttp_out (int32 c);

/* PTR data structures

   ptr_dev      PTR device descriptor
   ptr_unit     PTR unit descriptor
   ptr_mod      PTR modifiers
   ptr_reg      PTR register list
*/

DIB ptr_dib = { PTR, 0, 0, 0, 0, 0, &ptrio };

UNIT ptr_unit = {
    UDATA (&ptr_svc, UNIT_SEQ+UNIT_ATTABLE+UNIT_ROABLE, 0),
           SERIAL_IN_WAIT
    };

REG ptr_reg[] = {
    { ORDATA (BUF, ptr_unit.buf, 8) },
    { FLDATA (CMD, ptr_dib.cmd, 0) },
    { FLDATA (CTL, ptr_dib.ctl, 0) },
    { FLDATA (FLG, ptr_dib.flg, 0) },
    { FLDATA (FBF, ptr_dib.fbf, 0) },
    { FLDATA (SRQ, ptr_dib.srq, 0) },
    { DRDATA (TRLCTR, ptr_trlcnt, 8), REG_HRO },
    { DRDATA (TRLLIM, ptr_trllim, 8), PV_LEFT },
    { DRDATA (POS, ptr_unit.pos, T_ADDR_W), PV_LEFT },
    { DRDATA (TIME, ptr_unit.wait, 24), PV_LEFT },
    { FLDATA (STOP_IOE, ptr_stopioe, 0) },
    { ORDATA (DEVNO, ptr_dib.devno, 6), REG_HRO },
    { NULL }
    };

MTAB ptr_mod[] = {
    { UNIT_DIAG, UNIT_DIAG, "diagnostic mode", "DIAG", NULL },
    { UNIT_DIAG, 0, "reader mode", "READER", NULL },
    { MTAB_XTD | MTAB_VDV, 0, "DEVNO", "DEVNO",
      &hp_setdev, &hp_showdev, &ptr_dev },
    { 0 }
    };

DEVICE ptr_dev = {
    "PTR", &ptr_unit, ptr_reg, ptr_mod,
    1, 10, 31, 1, 8, 8,
    NULL, NULL, &ptr_reset,
    &ptr_boot, &ptr_attach, NULL,
    &ptr_dib, DEV_DISABLE
    };

/* PTP data structures

   ptp_dev      PTP device descriptor
   ptp_unit     PTP unit descriptor
   ptp_mod      PTP modifiers
   ptp_reg      PTP register list
*/

DIB ptp_dib = { PTP, 0, 0, 0, 0, 0, &ptpio };

UNIT ptp_unit = {
    UDATA (&ptp_svc, UNIT_SEQ+UNIT_ATTABLE, 0), SERIAL_OUT_WAIT
    };

REG ptp_reg[] = {
    { ORDATA (BUF, ptp_unit.buf, 8) },
    { FLDATA (CMD, ptp_dib.cmd, 0) },
    { FLDATA (CTL, ptp_dib.ctl, 0) },
    { FLDATA (FLG, ptp_dib.flg, 0) },
    { FLDATA (FBF, ptp_dib.fbf, 0) },
    { FLDATA (SRQ, ptp_dib.srq, 0) },
    { DRDATA (POS, ptp_unit.pos, T_ADDR_W), PV_LEFT },
    { DRDATA (TIME, ptp_unit.wait, 24), PV_LEFT },
    { FLDATA (STOP_IOE, ptp_stopioe, 0) },
    { ORDATA (DEVNO, ptp_dib.devno, 6), REG_HRO },
    { NULL }
    };

MTAB ptp_mod[] = {
    { MTAB_XTD | MTAB_VDV, 0, "DEVNO", "DEVNO",
      &hp_setdev, &hp_showdev, &ptp_dev },
    { 0 }
    };

DEVICE ptp_dev = {
    "PTP", &ptp_unit, ptp_reg, ptp_mod,
    1, 10, 31, 1, 8, 8,
    NULL, NULL, &ptp_reset,
    NULL, NULL, NULL,
    &ptp_dib, DEV_DISABLE
    };

/* TTY data structures

   tty_dev      TTY device descriptor
   tty_unit     TTY unit descriptor
   tty_reg      TTY register list
   tty_mod      TTy modifiers list
*/

#define TTI     0
#define TTO     1
#define TTP     2

DIB tty_dib = { TTY, 0, 0, 0, 0, 0, &ttyio };

UNIT tty_unit[] = {
    { UDATA (&tti_svc, UNIT_IDLE | TT_MODE_UC, 0), POLL_WAIT },
    { UDATA (&tto_svc, TT_MODE_UC, 0), TTY_OUT_WAIT },
    { UDATA (&tto_svc, UNIT_SEQ | UNIT_ATTABLE | TT_MODE_8B, 0), SERIAL_OUT_WAIT }
    };

REG tty_reg[] = {
    { ORDATA (BUF, tty_buf, 8) },
    { ORDATA (MODE, tty_mode, 16) },
    { ORDATA (SHIN, tty_shin, 8), REG_HRO },
    { FLDATA (CMD, tty_dib.cmd, 0), REG_HRO },
    { FLDATA (CTL, tty_dib.ctl, 0) },
    { FLDATA (FLG, tty_dib.flg, 0) },
    { FLDATA (FBF, tty_dib.fbf, 0) },
    { FLDATA (SRQ, tty_dib.srq, 0) },
    { FLDATA (KLFP, tty_lf, 0), REG_HRO },
    { DRDATA (KPOS, tty_unit[TTI].pos, T_ADDR_W), PV_LEFT },
    { DRDATA (KTIME, tty_unit[TTI].wait, 24), REG_NZ + PV_LEFT },
    { DRDATA (TPOS, tty_unit[TTO].pos, T_ADDR_W), PV_LEFT },
    { DRDATA (TTIME, tty_unit[TTO].wait, 24), REG_NZ + PV_LEFT },
    { DRDATA (PPOS, tty_unit[TTP].pos, T_ADDR_W), PV_LEFT },
    { FLDATA (STOP_IOE, ttp_stopioe, 0) },
    { ORDATA (DEVNO, tty_dib.devno, 6), REG_HRO },
    { NULL }
    };

MTAB tty_mod[] = {
    { TT_MODE, TT_MODE_UC, "UC", "UC", &tty_set_opt },
    { TT_MODE, TT_MODE_7B, "7b", "7B", &tty_set_opt },
    { TT_MODE, TT_MODE_8B, "8b", "8B", &tty_set_opt },
    { TT_MODE, TT_MODE_7P, "7p", "7P", &tty_set_opt },
    { UNIT_AUTOLF, UNIT_AUTOLF, "autolf", "AUTOLF", &tty_set_alf },
    { UNIT_AUTOLF, 0          , NULL, "NOAUTOLF", &tty_set_alf },
/* added by WTN */
    { UNIT_TRANSLATE, UNIT_TRANSLATE, "translate HP to VT100", "TRANSLATE", NULL },
    { UNIT_TRANSLATE, 0, NULL, "NOTRANSLATE", NULL },
/* end WTN switches */
    { MTAB_XTD | MTAB_VDV, 0, "DEVNO", "DEVNO",
      &hp_setdev, &hp_showdev, &tty_dev },
    { 0 }
    };

DEVICE tty_dev = {
    "TTY", tty_unit, tty_reg, tty_mod,
    3, 10, 31, 1, 8, 8,
    NULL, NULL, &tty_reset,
    NULL, NULL, NULL,
    &tty_dib, 0
    };

/* CLK data structures

   clk_dev      CLK device descriptor
   clk_unit     CLK unit descriptor
   clk_mod      CLK modifiers
   clk_reg      CLK register list
*/

DIB clk_dib = { CLK, 0, 0, 0, 0, 0, &clkio };

UNIT clk_unit = { UDATA (&clk_svc, UNIT_IDLE, 0) };

REG clk_reg[] = {
    { ORDATA (SEL, clk_select, 3) },
    { DRDATA (CTR, clk_ctr, 14) },
    { FLDATA (CMD, clk_dib.cmd, 0), REG_HRO },
    { FLDATA (CTL, clk_dib.ctl, 0) },
    { FLDATA (FLG, clk_dib.flg, 0) },
    { FLDATA (FBF, clk_dib.fbf, 0) },
    { FLDATA (SRQ, clk_dib.srq, 0) },
    { FLDATA (ERR, clk_error, CLK_V_ERROR) },
    { BRDATA (TIME, clk_time, 10, 24, 8) },
    { DRDATA (IPTICK, clk_tick, 24), PV_RSPC | REG_RO },
    { ORDATA (DEVNO, clk_dib.devno, 6), REG_HRO },
    { NULL }
    };

MTAB clk_mod[] = {
    { UNIT_DIAG, UNIT_DIAG, "diagnostic mode", "DIAG", NULL },
    { UNIT_DIAG, 0, "calibrated", "CALIBRATED", NULL },
    { MTAB_XTD | MTAB_VDV, 0, "DEVNO", "DEVNO",
      &hp_setdev, &hp_showdev, &clk_dev },
    { 0 }
    };

DEVICE clk_dev = {
    "CLK", &clk_unit, clk_reg, clk_mod,
    1, 0, 0, 0, 0, 0,
    NULL, NULL, &clk_reset,
    NULL, NULL, NULL,
    &clk_dib, DEV_DISABLE
    };

/* Paper tape reader IO instructions */

int32 ptrio (int32 inst, int32 IR, int32 dat)
{
int32 dev;

dev = IR & I_DEVMASK;                                   /* get device no */
switch (inst) {                                         /* case on opcode */

    case ioFLG:                                         /* flag clear/set */
        if ((IR & I_HC) == 0) { setFSR (dev); }         /* STF */
        break;

    case ioSFC:                                         /* skip flag clear */
        if (FLG (dev) == 0) PC = (PC + 1) & VAMASK;
        break;

    case ioSFS:                                         /* skip flag set */
        if (FLG (dev) != 0) PC = (PC + 1) & VAMASK;
        break;

    case ioMIX:                                         /* merge */
        dat = dat | ptr_unit.buf;
        break;

    case ioLIX:                                         /* load */
        dat = ptr_unit.buf;
        break;

    case ioCRS:                                         /* control reset */
                                                        /* action same as CLC */
    case ioCTL:                                         /* control clear/set */
        if (IR & I_CTL) {                               /* CLC */
            clrCMD (dev);                               /* clear cmd, ctl */
            clrCTL (dev);
            }
        else {                                          /* STC */
            setCMD (dev);                               /* set cmd, ctl */
            setCTL (dev);
            sim_activate (&ptr_unit, ptr_unit.wait);
            }
        break;

    default:
        break;
        }

if (IR & I_HC) { clrFSR (dev); }                        /* H/C option */
return dat;
}

/* Unit service */

t_stat ptr_svc (UNIT *uptr)
{
int32 dev, temp;

dev = ptr_dib.devno;                                    /* get device no */
clrCMD (dev);                                           /* clear cmd */
if ((ptr_unit.flags & UNIT_ATT) == 0)                   /* attached? */
    return IORETURN (ptr_stopioe, SCPE_UNATT);
while ((temp = getc (ptr_unit.fileref)) == EOF) {       /* read byte, error? */
    if (feof (ptr_unit.fileref)) {                      /* end of file? */
        if ((ptr_unit.flags & UNIT_DIAG) && (ptr_unit.pos > 0)) {
            rewind (ptr_unit.fileref);                  /* rewind if loop mode */
            ptr_unit.pos = 0;
            }
        else {
            if (ptr_trlcnt >= ptr_trllim) {             /* added all trailer? */
                if (ptr_stopioe) {                      /* stop on error? */
                    printf ("PTR end of file\n");
                    return SCPE_IOERR;
                    }
                else return SCPE_OK;                    /* no, just hang */
                }
            ptr_trlcnt++;                               /* count trailer */
            temp = 0;                                   /* read a zero */
            break;
            }
        }
    else {                                              /* no, real error */
        perror ("PTR I/O error");
        clearerr (ptr_unit.fileref);
        return SCPE_IOERR;
        }
    }
setFSR (dev);                                           /* set flag */
ptr_unit.buf = temp & 0377;                             /* put byte in buf */
ptr_unit.pos = ftell (ptr_unit.fileref);

if (temp)                                               /* character non-null? */
    ptr_trlcnt = 0;                                     /* clear trailing null counter */

return SCPE_OK;
}

/* Attach routine - clear the trailer counter */

t_stat ptr_attach (UNIT *uptr, char *cptr)
{
ptr_trlcnt = 0;
return attach_unit (uptr, cptr);
}

/* Reset routine - called from SCP, flags in DIB's */

t_stat ptr_reset (DEVICE *dptr)
{
ptr_dib.cmd = ptr_dib.ctl = 0;                          /* clear cmd, ctl */
ptr_dib.flg = ptr_dib.fbf = ptr_dib.srq = 1;            /* set flg, fbf, srq */
ptr_unit.buf = 0;
sim_cancel (&ptr_unit);                                 /* deactivate unit */
return SCPE_OK;
}

/* Paper tape reader bootstrap routine (HP 12992K ROM) */

const uint16 ptr_rom[IBL_LNT] = {
    0107700,                    /*ST CLC 0,C            ; intr off */
    0002401,                    /*   CLA,RSS            ; skip in */
    0063756,                    /*CN LDA M11            ; feed frame */
    0006700,                    /*   CLB,CCE            ; set E to rd byte */
    0017742,                    /*   JSB READ           ; get #char */
    0007306,                    /*   CMB,CCE,INB,SZB    ; 2's comp */
    0027713,                    /*   JMP *+5            ; non-zero byte */
    0002006,                    /*   INA,SZA            ; feed frame ctr */
    0027703,                    /*   JMP *-3 */
    0102077,                    /*   HLT 77B            ; stop */
    0027700,                    /*   JMP ST             ; next */
    0077754,                    /*   STA WC             ; word in rec */
    0017742,                    /*   JSB READ           ; get feed frame */
    0017742,                    /*   JSB READ           ; get address */
    0074000,                    /*   STB 0              ; init csum */
    0077755,                    /*   STB AD             ; save addr */
    0067755,                    /*CK LDB AD             ; check addr */
    0047777,                    /*   ADB MAXAD          ; below loader */
    0002040,                    /*   SEZ                ; E =0 => OK */
    0027740,                    /*   JMP H55 */
    0017742,                    /*   JSB READ           ; get word */
    0040001,                    /*   ADA 1              ; cont checksum */
    0177755,                    /*   STA AD,I           ; store word */
    0037755,                    /*   ISZ AD */
    0000040,                    /*   CLE                ; force wd read */
    0037754,                    /*   ISZ WC             ; block done? */
    0027720,                    /*   JMP CK             ; no */
    0017742,                    /*   JSB READ           ; get checksum */
    0054000,                    /*   CPB 0              ; ok? */
    0027702,                    /*   JMP CN             ; next block */
    0102011,                    /*   HLT 11             ; bad csum */
    0027700,                    /*   JMP ST             ; next */
    0102055,                    /*H55 HALT 55           ; bad address */
    0027700,                    /*   JMP ST             ; next */
    0000000,                    /*RD 0 */
    0006600,                    /*   CLB,CME            ; E reg byte ptr */
    0103710,                    /*   STC RDR,C          ; start reader */
    0102310,                    /*   SFS RDR            ; wait */
    0027745,                    /*   JMP *-1 */
    0106410,                    /*   MIB RDR            ; get byte */
    0002041,                    /*   SEZ,RSS            ; E set? */
    0127742,                    /*   JMP RD,I           ; no, done */
    0005767,                    /*   BLF,CLE,BLF        ; shift byte */
    0027744,                    /*   JMP RD+2           ; again */
    0000000,                    /*WC 000000             ; word count */
    0000000,                    /*AD 000000             ; address */
    0177765,                    /*M11 -11               ; feed count */
    0, 0, 0, 0, 0, 0, 0, 0,     /* unused */
    0, 0, 0, 0, 0, 0, 0,        /* unused */
    0000000                     /*MAXAD -ST             ; max addr */
    };

t_stat ptr_boot (int32 unitno, DEVICE *dptr)
{
int32 dev;

dev = ptr_dib.devno;                                    /* get device no */
if (ibl_copy (ptr_rom, dev)) return SCPE_IERR;          /* copy boot to memory */
SR = (SR & IBL_OPT) | IBL_PTR | (dev << IBL_V_DEV);     /* set SR */
return SCPE_OK;
}

/* Paper tape punch IO instructions */

int32 ptpio (int32 inst, int32 IR, int32 dat)
{
int32 dev;

dev = IR & I_DEVMASK;                                   /* get device no */
switch (inst) {                                         /* case on opcode */

    case ioFLG:                                         /* flag clear/set */
        if ((IR & I_HC) == 0) { setFSR (dev); }         /* STF */
        break;

    case ioSFC:                                         /* skip flag clear */
        if (FLG (dev) == 0) PC = (PC + 1) & VAMASK;
        break;

    case ioSFS:                                         /* skip flag set */
        if (FLG (dev) != 0) PC = (PC + 1) & VAMASK;
        break;

    case ioLIX:                                         /* load */
        dat = 0;
    case ioMIX:                                         /* merge */
        if ((ptp_unit.flags & UNIT_ATT) == 0)
            dat = dat | PTP_LOW;                        /* out of tape? */
        break;

    case ioOTX:                                         /* output */
        ptp_unit.buf = dat;
        break;

    case ioCRS:                                         /* control reset */
                                                        /* action same as CLC */
    case ioCTL:                                         /* control clear/set */
        if (IR & I_CTL) {                               /* CLC */
            clrCMD (dev);                               /* clear cmd, ctl */
            clrCTL (dev);
            }
        else {                                          /* STC */
            setCMD (dev);                               /* set cmd, ctl */
            setCTL (dev);
            sim_activate (&ptp_unit, ptp_unit.wait);
            }
        break;

    default:
        break;
        }

if (IR & I_HC) { clrFSR (dev); }                        /* H/C option */
return dat;
}

/* Unit service */

t_stat ptp_svc (UNIT *uptr)
{
int32 dev;

dev = ptp_dib.devno;                                    /* get device no */
clrCMD (dev);                                           /* clear cmd */
setFSR (dev);                                           /* set flag */
if ((ptp_unit.flags & UNIT_ATT) == 0)                   /* attached? */
    return IORETURN (ptp_stopioe, SCPE_UNATT);
if (putc (ptp_unit.buf, ptp_unit.fileref) == EOF) {     /* output byte */
    perror ("PTP I/O error");
    clearerr (ptp_unit.fileref);
    return SCPE_IOERR;
    }
ptp_unit.pos = ftell (ptp_unit.fileref);                /* update position */
return SCPE_OK;
}

/* Reset routine */

t_stat ptp_reset (DEVICE *dptr)
{
ptp_dib.cmd = ptp_dib.ctl = 0;                          /* clear cmd, ctl */
ptp_dib.flg = ptp_dib.fbf = ptp_dib.srq = 1;            /* set flg, fbf, srq */
ptp_unit.buf = 0;
sim_cancel (&ptp_unit);                                 /* deactivate unit */
return SCPE_OK;
}

/* Terminal IO instructions */

int32 ttyio (int32 inst, int32 IR, int32 dat)
{
int32 dev;

dev = IR & I_DEVMASK;                                   /* get device no */
switch (inst) {                                         /* case on opcode */

    case ioFLG:                                         /* flag clear/set */
        if ((IR & I_HC) == 0) { setFSR (dev); }         /* STF */
        break;

    case ioSFC:                                         /* skip flag clear */
        if (FLG (dev) == 0) PC = (PC + 1) & VAMASK;
        break;

    case ioSFS:                                         /* skip flag set */
        if (FLG (dev) != 0) PC = (PC + 1) & VAMASK;
        break;

    case ioLIX:                                         /* load */
        dat = 0;
    case ioMIX:                                         /* merge */
        dat = dat | tty_buf;
        if (!(tty_mode & TM_KBD) && sim_is_active (&tty_unit[TTO]))
            dat = dat | TP_BUSY;
        break;

    case ioOTX:                                         /* output */
        if (dat & TM_MODE) tty_mode = dat & (TM_KBD|TM_PRI|TM_PUN);
        tty_buf = dat & 0377;
        break;

    case ioCRS:                                         /* control reset */
        clrCTL (dev);                                   /* clear control */
        setFSR (dev);                                   /* set flag */
        tty_mode = TM_KBD;                              /* set tty, clear print/punch */
        tty_buf = 0;                                    /* clear buffer */
        tty_shin = 0377;                                /* input inactive */
        tty_lf = 0;                                     /* no lf pending */
        break;

    case ioCTL:                                         /* control clear/set */
        if (IR & I_CTL) { clrCTL (dev); }               /* CLC */
        else {                                          /* STC */
            setCTL (dev);
            if (!(tty_mode & TM_KBD))                   /* output? */
                sim_activate (&tty_unit[TTO], tty_unit[TTO].wait);
            }
        break;

    default:
        break;
        }

if (IR & I_HC) { clrFSR (dev); }                        /* H/C option */
return dat;
}

/* TTY input service routine.

   The console input poll routine is scheduled with a ten millisecond period
   using a calibrated timer, which is the source of event timing for all of the
   keyboard polling routines.  Synchronizing other keyboard polls with the
   console poll ensures maximum idle time.

   Several HP operating systems require a CR and LF sequence for line
   termination.  This is awkward on a PC, as there is no LF key (CTRL+J is
   needed instead).  We provide an AUTOLF mode to add a LF automatically to each
   CR input.  When this mode is set, entering CR will set a flag, which will
   cause a LF to be supplied automatically at the next input poll.

   The 12531C teleprinter interface and the later 12880A CRT interface provide a
   clever mechanism to detect a keypress during output.  This is used by DOS and
   RTE to allow the user to interrupt lengthy output operations to enter system
   commands.

   Referring to the 12531C schematic, the terminal input enters on pin X
   ("DATA FROM EIA COMPATIBLE DEVICE").  The signal passes through four
   transistor inversions (Q8, Q1, Q2, and Q3) to appear on pin 12 of NAND gate
   U104C.  If the flag flip-flop is not set, the terminal input passes to the
   (inverted) output of U104C and thence to the D input of the first of the
   flip-flops forming the data register.

   In the idle condition (no key pressed), the terminal input line is marking
   (voltage negative), so in passing through a total of five inversions, a
   logic one is presented at the serial input of the data register.  During an
   output operation, the register is parallel loaded and serially shifted,
   sending the output data through the register to the device and -- this is
   the crux -- filling the register with logic ones from U104C.

   At the end of the output operation, the card flag is set, an interrupt
   occurs, and the RTE driver is entered.  The driver then does an LIA SC to
   read the contents of the data register.  If no key has been pressed during
   the output operation, the register will read as all ones (octal 377).  If,
   however, any key was struck, at least one zero bit will be present.  If the
   register value doesn't equal 377, the driver sets the system "operator
   attention" flag, which will cause DOS or RTE to output an asterisk prompt and
   initiate a terminal read when the current output line is completed.
*/

t_stat tti_svc (UNIT *uptr)
{
int32 c, dev;

uptr->wait = sim_rtcn_calb (POLL_RATE, TMR_POLL);       /* calibrate poll timer */
sim_activate (uptr, uptr->wait);                        /* continue poll */

dev = tty_dib.devno;                                    /* get device no */
tty_shin = 0377;                                        /* assume inactive */
if (tty_lf) {                                           /* auto lf pending? */
    c = 012;                                            /* force lf */
    tty_lf = 0;
    }
else {
    if ((c = sim_poll_kbd ()) < SCPE_KFLAG) return c;   /* no char or error? */
    if (c & SCPE_BREAK) c = 0;                          /* break? */
    else c = sim_tt_inpcvt (c, TT_GET_MODE (uptr->flags));
    tty_lf = ((c & 0177) == 015) && (uptr->flags & UNIT_AUTOLF);
    }
if (tty_mode & TM_KBD) {                                /* keyboard enabled? */
    tty_buf = c;                                        /* put char in buf */
    uptr->pos = uptr->pos + 1;
    setFSR (dev);                                       /* set flag */
    if (c) {
        tto_out (c);                                    /* echo? */
        return ttp_out (c);                             /* punch? */
        }
    }
else tty_shin = c;                                      /* no, char shifts in */
return SCPE_OK;
}

/* TTY output service routine */

t_stat tto_svc (UNIT *uptr)
{
int32 c, dev;
t_stat r;

c = tty_buf;                                            /* get char */
tty_buf = tty_shin;                                     /* shift in */
tty_shin = 0377;                                        /* line inactive */
if ((r = tto_out (c)) != SCPE_OK) {                     /* output; error? */
    sim_activate (uptr, uptr->wait);                    /* retry */
    return ((r == SCPE_STALL)? SCPE_OK: r);             /* !stall? report */
    }
dev = tty_dib.devno;                                    /* get device no */
setFSR (dev);                                           /* set done flag */
return ttp_out (c);                                     /* punch if enabled */
}

t_stat tto_out (int32 c)
{
t_stat r;

if (tty_mode & TM_PRI) {                                /* printing? */
    c = sim_tt_outcvt (c, TT_GET_MODE (tty_unit[TTO].flags));
    if (c >= 0) {                                       /* valid? */
/* begin WTN code mods */
       if ((tty_unit[TTO].flags & UNIT_TRANSLATE) == 0) { /* if not enabled do original code */
         if (r = sim_putchar_s (c)) return r;            /* output char */
         tty_unit[TTO].pos = tty_unit[TTO].pos + 1;
       } else { /* otherwise translate HP codes to VT100 codes... */

/*************************************************************************/
/* Partial HP2615 / HP2621 / HP2645 emulation 8/8/05 by WTN
/* uses globals pnormal escflag escstate cancelesc escchar1 thru escchar3
/* calls sim_putchar to output additional characters
/* entry: c containing character
/* exit: c unchanged, pnormal=0 to supress normal printing
/* Supported translations...
/* <esc>A - up one line - outputs <esc>[0m<esc>[A
/* <esc>B - down one line - outputs <esc>[0m<esc>[B
/* <esc>C - right one char - outputs <esc>[C
/* <esc>D - left one char - outputs <esc>[D
/* <esc>E - clears screen and homes - outputs <esc>[0m<esc>[2J<esc>[1;1H
/* <esc>F - positions at line 22 column 1 - outputs <esc>[0m<esc>[22;1H
/* <esc>G - positions at beginning of current line - outputs <esc>[79D
/* <esc>H - positions at line 1 column 1 (home) - outputs <esc>[0m<esc>[1;1H
/* <esc>h - same as esc H
/* <esc>I - outputs a tab character, moving to next tab position
/* <esc>i - outputs a tab then <esc>[16D, semi-simulating a back-tab
/* <esc>J - clears display from current position - outputs <esc>[0m<esc>[J
/* <esc>K - clears line from current position - outputs <esc>[K
/* Cursor movement... note esc&a addressing begins at 0,0
/* <esc>&a20r40C or <esc>&a40c20R sets cursor to line 21 column 41
/* <esc>&a-5r+10C moves cursor up 5 lines and right 10 columns
/* <esc>&a+10C or <esc>&a+0r+10C move cursor right 10 columns on current line
/* <esc>&a20y40C or <esc>&a40c20Y is the same as <esc>&a20r40C
/* Note... unlike the HP2645 the cursor does not wrap! Neither do PC terms.
/* Implementing scrolling and wrap or even a real <esc>F would involve
/* way too much stuff to even think about putting it in SIMH, save it
/* for a custom Telnet program with a HP2645 emulation setting, there you
/* have line buffers, the VT/ANSI engine and all the other needed stuff.
/* This code translates each part of a <esc>&a sequence separately,
/* inserting additional ANSI escapes until an uppercase letter received.
/* Embedded spaces can appear in the numeric part of the sequence.
/* Absolute #c/C codes output <esc>[79D then if #>0 <esc>#C
/* Absolute #r/R/y/Y codes output <esc>[0m[24A then if #>0 <esc>#B
/* Relative codes output <esc>[#A B C or D depending on direction.
/* Relative codes that go nowhere (-0r etc) output <esc><space> (nop)
/* Relative codes that move off row output <esc>[0m first.
/* Attributes... note some of these are different from a real HP2645
/* substituted bold for some of the blinks (can do without the blinking)
/* <esc>&d@ - cancel attributes       outputs <esc>[0m
/* <esc>&dA - bold (for blink)        outputs <esc>[0m<esc>[1m
/* <esc>&dB - reverse                 outputs <esc>[0m<esc>[7m
/* <esc>&dC - noblink reverse         outputs <esc>[0m<esc>[7m
/* <esc>&dD - underline               outputs <esc>[0m<esc>[4m
/* <esc>&dE - underline bold          outputs <esc>[0m<esc>[4;1m
/* <esc>&dF - underline reverse       outputs <esc>[0m<esc>[4;7m
/* <esc>&dG - underln noblink reverse outputs <esc>[0m<esc>[4;7m
/* <esc>&dH - faint                   outputs <esc>[0m<esc>[2m
/* <esc>&dI - faint blink             outputs <esc>[0m<esc>[2;6m
/* <esc>&dJ - faint reverse           outputs <esc>[0m<esc>[2;7m
/* <esc>&dK - faint blink reverse     outputs <esc>[0m<esc>[2;6;7m
/* <esc>&dL - faint underline         outputs <esc>[0m<esc>[2;4m
/* <esc>&dM - faint underline blink   outputs <esc>[0m<esc>[2;4;6m
/* <esc>&dN - faint underline reverse outputs <esc>[0m<esc>[2;4;7m
/* <esc>&dO - faint underln blink rev outputs <esc>[0m<esc>[2;4;6;7m
/* Note... attributes only apply to characters printed afterwards.
/* They cannot be used to make say a permanently reversed field.
/*
/* Status so far...
/* Existing text and VT/ANSI sequences at use [ are not affected.
/* STAR works ok if HP2615 mode is selected.
/* OTHEL (2 different versions) works ok, say NO to line-drawing.
/* WOODY works ok in both default HP2645 and /D2 HP2621 modes.
/* Under Win95's Telnet, home ops go to beginning of scroll buffer,
/*   clear screen (hit esc E) first to avoid losing cursor.
/* This doesn't happen under Linux's Telnet, homes to visible home.
/* XP's stock Telnet doesn't work, PuTTYtel works ok (now), set BS to 8.
/* Forget about using with a Windows ANSI.SYS that doesn't implement the
/*  <esc>[J clear from cursor command, mine clears the entire screen.
/*
/* Changes...
/* 2/4/07 - added code to cancel attribs after crlf if by esc& code
/* 8/8/05 - if null received after escape cancels sequence, nulls removed
/*          made cancel character adjustable, space doesn't always work
/* 8/7/05 - modified attribs, tweeked code and comments
/* 8/6/05 - initial sharing
/*************************************************************************/
	pnormal = 1;                    /* by default print chars */
	if (escstate > 0) {             /* state-machine processing */
	    pnormal = 0;
	    if (escstate == 1) {        /* start esc& processing  */
		escchar1 = 0; escchar2 = 0; escchar3 = 0;
		if (c == 97) {              /* a */
		    escstate = 2;           /* switch to cursor control */
		} else if (c == 100) {      /* d */
		    escstate = 3;           /* switch to attribute control */
		} else if (c == 32 || c == 0) { /* ignore spaces */
		} else {                    /* unknown code */
		    escstate = 0;           /* terminate esc& processing */
		    sim_putchar (cancelesc);     /* cancel esc sequence */
		}
	    } else if (escstate == 2) {     /* cursor control... */
		if (c == 114 || c == 82 || c == 121 || c == 89) {
		    /* r or R or y or Y to set row position */
		    if (escchar1 == 43) {    /* + relative */
			if (escchar2 == 0) sim_putchar (' ');
			else {
			    sim_putchar ('[');
			    sim_putchar ('0'); /* cancel attribs */
			    sim_putchar ('m'); /* when moving to another line */
			    sim_putchar (27);
			    sim_putchar ('[');
			    sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('B');
			}
		    } else if (escchar1 == 45) { /* - relative */
			if (escchar2 == 0) sim_putchar (' ');
			else {
			    sim_putchar ('[');
			    sim_putchar ('0'); /* cancel attribs */
			    sim_putchar ('m'); /* when moving to another line */
			    sim_putchar (27);
			    sim_putchar ('[');
			    sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('A');
			}
		    } else { /* absolute */
			sim_putchar ('[');
			sim_putchar ('0'); /* cancel attribs */
			sim_putchar ('m'); /* when moving to another line */
			sim_putchar (27);
			sim_putchar ('[');
			sim_putchar ('2');
			sim_putchar ('4');
			sim_putchar ('A');
			if (escchar1 != 0) {  /* if row 0 stay put */
			    sim_putchar (27);
			    sim_putchar ('[');
			    sim_putchar (escchar1);
			    if (escchar2 != 0) sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('B');
			}
		    }
		    if (c < 97) escstate = 0; /* UC code exits sequence */
		    else { escchar1 = 0; escchar2 = 0;
			   escchar3 = 0; sim_putchar (27); }
		} else if (c == 99 || c == 67) {
		    /* c or C to set column position */
		    if (escchar1 == 43) {    /* + relative */
			if (escchar2 == 0) sim_putchar (' ');
			else {
			    sim_putchar ('[');
			    sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('C');
			}
		    } else if (escchar1 == 45) { /* - relative */
			if (escchar2 == 0) sim_putchar (' ');
			else {
			    sim_putchar ('[');
			    sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('D');
			}
		    } else { /* absolute */
			sim_putchar ('[');
			sim_putchar ('7');
			sim_putchar ('9');
			sim_putchar ('D');
			if (escchar1 != 0) { /* if col 0 stay put */
			    sim_putchar (27);
			    sim_putchar ('[');
			    sim_putchar (escchar1);
			    if (escchar2 != 0) sim_putchar (escchar2);
			    if (escchar3 != 0) sim_putchar (escchar3);
			    sim_putchar ('C');
			}
		    }
		    if (c < 97) escstate = 0; /* UC code exits sequence */
		    else { escchar1 = 0; escchar2 = 0;
			   escchar3 = 0; sim_putchar (27); }
		} else { /* must be a parm, suck it up */
		    if (c < 43 || c > 57); /* ignore garbage */
		    else if (escchar1 == 0 && c == 48); /* ignore leading zeros */
		    else if (escchar1 == 43 && escchar2 == 0 && c == 48); /* relative */
		    else if (escchar1 == 45 && escchar2 == 0 && c == 48); /* addr too */
		    else if (escchar1 == 0) escchar1 = c;
		    else if (escchar2 == 0) escchar2 = c;
		    else if (escchar3 == 0) escchar3 = c;
		    /* else escstate = 0; /* overflow */
		}
	    } else if (escstate == 3) {     /* attribute control... */
		/* HP2645 implements as a window which means I'd need a copy */
		/* of the screen which would take parallel interpreting of all */
		/* the codes... way too much stuff! so very limited support... */
		/* only works if chars to be enhanced are surrounded by codes */
		if (c != 32 && c != 0) { /* ignore embedded spaces */
		    sim_putchar ('['); /* start ansi sequence, esc already printed */
		    sim_putchar ('0');
		    sim_putchar ('m'); /* cancel previous attributes */
                    doingattrib = 0;
		    if (c != 64) { sim_putchar (27);    /* start new sequence */
				   sim_putchar ('['); } /* if not a cancel op */
		    if (c == 64) {         /* @ - cancel */
			/* do nothing, already cancelled */
		    } else if (c == 65) {  /* A - blink */
			sim_putchar ('1'); /* --> bold */
			doingattrib = 1;
		    } else if (c == 66) {  /* B - reverse  */
			sim_putchar ('7'); /* --> reverse */
			doingattrib = 1;
		    } else if (c == 67) {  /* C reverse blink */
			sim_putchar ('7'); /* --> reverse    */
			doingattrib = 1;
		    } else if (c == 68) {  /* D - underline  */
			sim_putchar ('4'); /* --> underline */
			doingattrib = 1;
		    } else if (c == 69) {  /* E - underline blink */
			sim_putchar ('4'); /* --> underline      */
			sim_putchar (';'); 
			sim_putchar ('1'); /* --> bold         */
			doingattrib = 1;
		    } else if (c == 70) {  /* F - underline reverse */
			sim_putchar ('4'); /* --> underline        */
			sim_putchar (';'); 
			sim_putchar ('7'); /* --> reverse        */
			doingattrib = 1;
		    } else if (c == 71) {  /* G - underline reverse blink */
			sim_putchar ('4'); /* --> underline              */
			sim_putchar (';');
			sim_putchar ('7'); /* --> reverse              */
			doingattrib = 1;
		    } else if (c == 72) {  /* H - dim    */
			sim_putchar ('2'); /* --> faint */
			doingattrib = 1;
		    } else if (c == 73) {  /* I - dim blink */
			sim_putchar ('2'); /* --> faint    */
			sim_putchar (';'); 
			sim_putchar ('6'); /* --> blink  */
			doingattrib = 1;
		    } else if (c == 74) {  /* J - dim reverse */
			sim_putchar ('2'); /* --> faint      */
			sim_putchar (';'); 
			sim_putchar ('7'); /* --> reverse  */
			doingattrib = 1;
		    } else if (c == 75) {  /* K - dim reverse blink */
			sim_putchar ('2'); /* --> faint            */
			sim_putchar (';'); 
			sim_putchar ('6'); /* --> blink          */
			sim_putchar (';');
			sim_putchar ('7'); /* --> reverse      */
			doingattrib = 1;
		    } else if (c == 76) {  /* L - dim underline */
			sim_putchar ('2'); /* --> faint        */
			sim_putchar (';'); 
			sim_putchar ('4'); /* --> underline  */
			doingattrib = 1;
		    } else if (c == 77) {  /* M - dim underline blink */
			sim_putchar ('2'); /* --> faint              */
			sim_putchar (';'); 
			sim_putchar ('4'); /* --> underline        */
			sim_putchar (';'); 
			sim_putchar ('6'); /* --> blink          */
			doingattrib = 1;
		    } else if (c == 78) {  /* N - dim underline reverse */
			sim_putchar ('2'); /* --> faint                */
			sim_putchar (';'); 
			sim_putchar ('4'); /* --> underline          */
			sim_putchar (';'); 
			sim_putchar ('7'); /* --> reverse          */
			doingattrib = 1;
		    } else if (c == 79) {  /* O - dim underline reverse blink */
			sim_putchar ('2'); /* --> faint                      */
			sim_putchar (';'); 
			sim_putchar ('4'); /* --> underline                */
			sim_putchar (';'); 
			sim_putchar ('6'); /* --> blink                  */
			sim_putchar (';');
			sim_putchar ('7'); /* --> reverse              */
			doingattrib = 1;
		    } else {
			sim_putchar ('0'); /* any other cancels attribs  */
			doingattrib = 0;
		    }
		    if (c != 64) sim_putchar ('m'); /* if not @ do sequence */
		    escstate = 0;       /* exit enhance mode */
		}
		pnormal = 0;        /* don't print char */
	    } else escstate = 0;    /* invalid state shouldn't happen */
	} else if (escflag != 0) {  /* esc received last time      */
	    escflag = 0;
	    /* fix attribs, specs say only affect current line */
	    if (c == 65 || c == 66 || c == 69 ||
		c == 70 || c == 72 || c == 74) {
		sim_putchar ('[');
		sim_putchar ('0');
		sim_putchar ('m');
		sim_putchar (27);
	    }
	    if (c == 38) {          /* & code */
		escstate = 1;       /* engage & interpreter next time */
		pnormal = 0;        /* don't print the & */
            } else if (c == 0) {
                sim_putchar (cancelesc); /* cancel escape */
                pnormal = 0;   /* don't pass to term */
	    } else if (c != 91) {   /* ignore if regular esc[ sequence */
		/* ignore if not recognized sequence... A-K h i */
		if ((c >= 65 && c <= 75) || c == 104 || c == 105 ) {
		    sim_putchar ('[');           /* make VT-style */
		    if (c == 69) {               /* E clear screen */
			sim_putchar ('2');
			sim_putchar ('J');
			sim_putchar (27);
			sim_putchar ('[');
			sim_putchar ('1');
			sim_putchar (';');
			sim_putchar ('1');
			sim_putchar ('H');
			pnormal = 0; /* don't print */
		    } else if (c == 72 || c == 104) {  /* H or h home up */
			sim_putchar ('1');
			sim_putchar (';');
			sim_putchar ('1');
			sim_putchar ('H');
			pnormal = 0; /* don't print */
		    } else if (c == 73) { /* I */
			sim_putchar (cancelesc); /* cancel current esc sequence */
			sim_putchar (9);   /* print a tab char */
			pnormal = 0; /* don't print */
		    } else if (c == 105) { /* i back tab */
			/* not perfect, do a tab then back up 16 positions */
			sim_putchar (cancelesc); /* cancel current esc sequence */
			sim_putchar (9);
			sim_putchar (27);
			sim_putchar ('[');
			sim_putchar ('1');
			sim_putchar ('6');
			sim_putchar ('D');
			pnormal = 0; /* don't print */
		    } else if (c == 70) { /* F home down */
			/* the 2645 docs say this is to position cursor after */
			/* the last line in display memory... don't know      */
			/* what line that is so...                            */
			sim_putchar ('2');
			sim_putchar ('2');
			sim_putchar (';');
			sim_putchar ('1');
			sim_putchar ('H');
			pnormal = 0; /* don't print */
		    } else if (c == 71) { /* G home line */
			sim_putchar ('7');
			sim_putchar ('9');
			sim_putchar ('D');
			pnormal = 0; /* don't print */ 
		    }  /* codes A B C D J K similar to VT/ANSI, pass thru */
		} /* else if (c==32) pnormal = 0; /* ignore spaces */
	    }
	}
        if (c == 0) pnormal = 0; /* strip out nulls in case xmit code doesn't */
	if (c == 27) {    /* if escape received at any time */
	    escflag = 1;  /* set escape flag */
	    escstate = 0; /* reset state machine */
	    sim_putchar (c); /* output escape directly */
	    pnormal = 0;  /* don't print esc in case being masked */
	}
        if (doingattrib!=0 && c==13) { /* if end of line */
            sim_putchar (27);
            sim_putchar ('[');  /* cancel prev line's attribs */
            sim_putchar ('0');
            sim_putchar ('m');
            doingattrib = 0;
        }
/**** end HP/VT translation, don't print c if pnormal=0 ****/
	if (pnormal != 0) {
          if (c >= 0) {                                     /* (still) valid? */
            if (r = sim_putchar_s (c)) return r;            /* output char   */
            tty_unit[TTO].pos = tty_unit[TTO].pos + 1;
            /* position won't be right when term graphics displayed */
          }
        }
      } /* endif enabled */
    }  /* endif valid */
  }   /* endif printing */
/* end WTN code */

return SCPE_OK;
}

t_stat ttp_out (int32 c)
{
if (tty_mode & TM_PUN) {                                /* punching? */
    if ((tty_unit[TTP].flags & UNIT_ATT) == 0)          /* attached? */
        return IORETURN (ttp_stopioe, SCPE_UNATT);
    if (putc (c, tty_unit[TTP].fileref) == EOF) {       /* output char */
        perror ("TTP I/O error");
        clearerr (tty_unit[TTP].fileref);
        return SCPE_IOERR;
        }
    tty_unit[TTP].pos = ftell (tty_unit[TTP].fileref);
    }
return SCPE_OK;
}

/* TTY reset routine */

t_stat tty_reset (DEVICE *dptr)
{
tty_dib.cmd = tty_dib.ctl = 0;                          /* clear cmd, ctl */
tty_dib.flg = tty_dib.fbf = tty_dib.srq = 1;            /* set flg, fbf, srq */
tty_mode = TM_KBD;                                      /* enable input */
tty_buf = 0;
tty_shin = 0377;                                        /* input inactive */
tty_lf = 0;                                             /* no lf pending */
tty_unit[TTI].wait = POLL_WAIT;                         /* reset initial poll */
sim_rtcn_init (tty_unit[TTI].wait, TMR_POLL);           /* init poll timer */
sim_activate (&tty_unit[TTI], tty_unit[TTI].wait);      /* activate poll */
sim_cancel (&tty_unit[TTO]);                            /* cancel output */
return SCPE_OK;
}

t_stat tty_set_opt (UNIT *uptr, int32 val, char *cptr, void *desc)
{
int32 u = uptr - tty_dev.units;

if (u > TTO) return SCPE_NOFNC;
if ((u == TTI) && (val == TT_MODE_7P))
    val = TT_MODE_7B;
tty_unit[u].flags = (tty_unit[u].flags & ~TT_MODE) | val;
return SCPE_OK;
}

t_stat tty_set_alf (UNIT *uptr, int32 val, char *cptr, void *desc)
{
int32 u = uptr - tty_dev.units;

if (u != TTI) return SCPE_NOFNC;
return SCPE_OK;
}

/* Synchronize polling.

   Return an event time corresponding either with the amount of time remaining
   in the current poll (mode = INITIAL) or the amount of time in a full poll
   period (mode = SERVICE).  If the former call is made when the device service
   routine is started, then making the latter call during unit service will
   ensure that the polls remain synchronized.
 */

int32 sync_poll (POLLMODE poll_mode)
{
int32 poll_time;

    if (poll_mode == INITIAL) {
        poll_time = sim_is_active (&tty_unit[TTI]);

        if (poll_time)
            return poll_time;
        else
            return POLL_WAIT;
        }
    else
        return tty_unit[TTI].wait;
}


/* Clock I/O instructions.

   The time base generator (CLK) provides periodic interrupts from 100
   microseconds to 1000 seconds.  The CLK uses a calibrated timer to provide the
   time base.  For periods ranging from 1 to 1000 seconds, a 100 millisecond
   timer is used, and 10 to 10000 ticks are counted before setting the device
   flag to indicate that the period has expired.

   If the period is set to ten milliseconds, the console poll timer is used
   instead of an independent timer.  This is to maximize the idle period.

   In diagnostic mode, the clock period is set to the expected number of CPU
   instructions, rather than wall-clock time, so that the diagnostic executes as
   expected.
*/

int32 clkio (int32 inst, int32 IR, int32 dat)
{
int32 dev;

dev = IR & I_DEVMASK;                                   /* get device no */
switch (inst) {                                         /* case on opcode */

    case ioFLG:                                         /* flag clear/set */
        if ((IR & I_HC) == 0) { setFSR (dev); }         /* STF */
        break;

    case ioSFC:                                         /* skip flag clear */
        if (FLG (dev) == 0) PC = (PC + 1) & VAMASK;
        break;

    case ioSFS:                                         /* skip flag set */
        if (FLG (dev) != 0) PC = (PC + 1) & VAMASK;
        break;

    case ioMIX:                                         /* merge */
        dat = dat | clk_error;
        break;

    case ioLIX:                                         /* load */
        dat = clk_error;
        break;

    case ioOTX:                                         /* output */
        clk_select = dat & 07;                          /* save select */
        sim_cancel (&clk_unit);                         /* stop the clock */
        clrCTL (dev);                                   /* clear control */
        break;

    case ioCRS:                                         /* control reset */
                                                        /* action same as CLC */
    case ioCTL:                                         /* control clear/set */
        if (IR & I_CTL) {                               /* CLC */
            clrCTL (dev);                               /* turn off clock */
            sim_cancel (&clk_unit);                     /* deactivate unit */
            }
        else {                                          /* STC */
            setCTL (dev);                               /* set CTL */

            if (clk_unit.flags & UNIT_DIAG)                     /* diag mode? */
                clk_unit.flags = clk_unit.flags & ~UNIT_IDLE;   /* not calibrated */
            else
                clk_unit.flags = clk_unit.flags | UNIT_IDLE;    /* is calibrated */

            if (!sim_is_active (&clk_unit)) {           /* clock running? */
                clk_tick = clk_delay (0);               /* get tick count */

                if ((clk_unit.flags & UNIT_DIAG) == 0)  /* calibrated? */
                    if (clk_select == 2)                /* 10 msec. interval? */
                        clk_tick = sync_poll (INITIAL); /* sync  poll */
                    else
                        sim_rtcn_init (clk_tick, TMR_CLK);  /* initialize timer */

                sim_activate (&clk_unit, clk_tick);     /* start clock */
                clk_ctr = clk_delay (1);                /* set repeat ctr */
                }
            clk_error = 0;                              /* clear error */
            }
        break;

    default:
        break;
        }

if (IR & I_HC) { clrFSR (dev); }                        /* H/C option */
return dat;
}

/* CLK unit service.

   As with the I/O handler, if the time base period is set to ten milliseconds,
   the console poll timer is used instead of an independent timer.
*/

t_stat clk_svc (UNIT *uptr)
{
int32 dev;

dev = clk_dib.devno;                                    /* get device no */
if (!CTL (dev)) return SCPE_OK;                         /* CTL off? done */

if (clk_unit.flags & UNIT_DIAG)                         /* diag mode? */
    clk_tick = clk_delay (0);                           /* get fixed delay */
else if (clk_select == 2)                               /* 10 msec period? */
    clk_tick = sync_poll (SERVICE);                     /* sync poll */
else
    clk_tick = sim_rtcn_calb (clk_tps[clk_select], TMR_CLK);    /* calibrate delay */

sim_activate (uptr, clk_tick);                          /* reactivate */
clk_ctr = clk_ctr - 1;                                  /* decrement counter */
if (clk_ctr <= 0) {                                     /* end of interval? */
    if (FLG (dev)) clk_error = CLK_ERROR;               /* overrun? error */
    else { setFSR (dev); }                              /* else set flag */
    clk_ctr = clk_delay (1);                            /* reset counter */
    }
return SCPE_OK;
}

/* Reset routine */

t_stat clk_reset (DEVICE *dptr)
{
clk_dib.cmd = clk_dib.ctl = 0;                          /* clear cmd, ctl */
clk_dib.flg = clk_dib.fbf = clk_dib.srq = 1;            /* set flg, fbf, srq */
clk_error = 0;                                          /* clear error */
clk_select = 0;                                         /* clear select */
clk_ctr = 0;                                            /* clear counter */
sim_cancel (&clk_unit);                                 /* deactivate unit */
return SCPE_OK;
}

/* Clock delay routine */

int32 clk_delay (int32 flg)
{
int32 sel = clk_select;

if ((clk_unit.flags & UNIT_DIAG) && (sel >= 4)) sel = sel - 3;
if (flg) return clk_rpt[sel];
else return clk_time[sel];
}
